// Copyright (c)2002-2003 Eric Carlson
// This source is licensed under the BSD license.
// Please see the LICENSE.txt file included with the distribution for more information.

#import "SliderController.h"

@implementation SliderController

#pragma mark -----------Sheet Handling Selectors--------------
//from Monitor->Revert Settings... menu item
- (IBAction)revertSettings:(id)sender
{
    NSBeginAlertSheet(@"Revert Settings", @"No", @"Yes", nil, mainWindow, self, NULL, @selector(revertSheetDidDismiss:returnCode:contextInfo:), nil, @"Are you sure you want to revert this monitor's settings for this mode?");
}

//called when user dismisses revert sheet
- (void)revertSheetDidDismiss:(NSWindow *)sheet
                   returnCode:(int)returnCode
                  contextInfo:(void *)contextInfo
{
    if(returnCode == NSAlertAlternateReturn) { //Yes was pressed
        [self raiseRevertProgressSheet];
        [myMonitor updateSettingsUsingArray:originalSettings
                       withProgressSelector:@selector(updateRevertProgress)
                           selectorWhenDone:@selector(endRevertProgressSheet)
                                    ofClass:self];
    }
}

//callback to update progress bar as revert is occurring
- (void)updateRevertProgress
{
    [revertProgress incrementBy:1.00];
    [revertProgress displayIfNeeded]; //make sure it's updated
}

//put up the revert progress sheet
- (void)raiseRevertProgressSheet
{
    [NSApp beginSheet:revertProgressSheet
       modalForWindow:mainWindow
        modalDelegate:self
       didEndSelector:@selector(revertProgressSheetDidEnd:returnCode:contextInfo:)
          contextInfo:nil];    
}

//callback passed to revert procedure to be called when reverting settings is finished
- (IBAction)endRevertProgressSheet
{
    //hide sheet
    [revertProgressSheet orderOut:self];

    //zero progress indicator
    [revertProgress setDoubleValue:0];

    //return to normal event handling
    [NSApp endSheet:revertProgressSheet returnCode:1];    
}

//called when revert progress sheet is done
- (void)revertProgressSheetDidEnd:(NSWindow *)sheet
                       returnCode:(int)returnCode
                      contextInfo:(void *)contextInfo
{
    [self updateControlValues];
}

//terminate the application when the user dismisses the no monitor alert sheet
- (void)noMonitorSheetDidEnd:(NSWindow *)sheet
                  returnCode:(int)returnCode
                 contextInfo:(void *)contextInfo
{
    [NSApp terminate:nil];
}


#pragma mark ----------------Action Targets ----------------
//called when the user changes the geometry control radio button
//change the tag associated with the slider and associated controls so that it controls the right thing
//also update the slider and stuff so it reflects the current settings in the monitor
- (IBAction)updateGeometrySliderTarget:(id)sender
{
    int selectedRow = [sender selectedRow];
    [geometrySlider setTag:geometryElementTags[selectedRow]];
    [geometryStepper setTag:geometryElementTags[selectedRow]];
    [geometryValue setTag:geometryElementTags[selectedRow]];

    [geometrySlider setFloatValue:[myMonitor getHIDElementValue:geometryElementTags[selectedRow]]];
    [geometryStepper setFloatValue:[myMonitor getHIDElementValue:geometryElementTags[selectedRow]]];
    [geometryValue setFloatValue:[myMonitor getHIDElementValue:geometryElementTags[selectedRow]]];
}

//called when degauss button is pressed
- (IBAction)degauss:(id)sender
{
    [myMonitor degauss];
}

//called when the current monitor is changed using the menu
- (void)changeTargetMonitor:(id)sender
{
    if([myMonitor selectedMonitor] != [sender tag]) {
        NSLog(@"Target changed.\n");
        [myMonitor changeSelectedMonitorTo:[sender tag]];
        [sender setState:NSOnState];
        [self updateControlValues];        
    }
}

//called when a slider, value, or stepper is changed
//updates the appropriate monitor control based on the control's tag value
- (void)changeValue:(id)sender
{
    int i;
    int foo;
    NSArray *tmp;
    NSNumber *bar;
    [myMonitor setHIDElement:[sender tag] withValue:[sender floatValue]];

    //update associated elements' values here
    foo = [sender tag];
    bar = [NSNumber numberWithInt:foo];
    tmp = [controlGroups objectForKey:bar];
    if(tmp != nil) {
        for(i=0; i<[tmp count]; i++) {
            [[tmp objectAtIndex:i] setFloatValue:[sender floatValue]];
        }        
    }
}

//take over the screen and show a test pattern, with the control window floating over that
- (void)showTestPatterns:(id)sender
{
    int windowLevel;
    NSRect screenRect;
    if(fullScreenWindow == nil) {
        // Capture the main display
        if (CGDisplayCapture( kCGDirectMainDisplay ) != kCGErrorSuccess) {
            NSLog( @"Couldn't capture the main display!" );
            // Note: you'll probably want to display a proper error dialog here
        }
        // Get the shielding window level
        windowLevel = CGShieldingWindowLevel();
        // Get the screen rect of our main display
        screenRect = [[NSScreen mainScreen] frame];
        // Put up a new window
        fullScreenWindow = [[NSWindow alloc] initWithContentRect:screenRect
                                                       styleMask:NSBorderlessWindowMask
                                                         backing:NSBackingStoreBuffered
                                                           defer:NO screen:[NSScreen mainScreen]];

        [fullScreenWindow setLevel:windowLevel];
        [fullScreenWindow setBackgroundColor:[NSColor blackColor]];
        patterns = [[TestPatternView alloc] initWithFrame:screenRect];
        [fullScreenWindow setContentView:patterns];
        [fullScreenWindow orderFront:self];

        [mainWindow setLevel:windowLevel];
        [mainWindow makeKeyAndOrderFront:self];        
    }
}

//release the full screen mode
- (void)hideTestPatterns:(id)sender
{
    if(fullScreenWindow != nil) {
        [fullScreenWindow orderOut:self];
        [fullScreenWindow dealloc];
        fullScreenWindow = nil;
        [mainWindow setLevel:NSNormalWindowLevel];

        // Release the display(s)
        if (CGDisplayRelease( kCGDirectMainDisplay ) != kCGErrorSuccess) {
            NSLog( @"Couldn't release the display(s)!" );
            // Note: if you display an error dialog here, make sure you set
            // its window level to the same one as the shield window level,
            // or the user won't see anything.
        }        
    }
}

//called by menu item
- (void)showControlsWindow:(id)sender
{
    if(![mainWindow isVisible]) {
        [mainWindow makeKeyAndOrderFront:self];
        [hideShowControlsWindowMenuItem setTitle:@"Hide Controls Window"];
    } else {
        [mainWindow performClose:self]; //calls windowShouldClose delegate method
    }
}

#pragma mark --------------Helper Methods------------------
//update the GUI controls to reflect current settings in the monitor
- (void)updateControlValues
{
    int i;
    id key;
    NSArray *tmp;
    NSEnumerator *enumerator = [controlGroups keyEnumerator];

    //update the tile to reflect the new display mode info
    [mainWindow setTitle:
        [NSString stringWithFormat:@"Diamond Control X: %ld x %ld @ %ld hz",current_width,current_height,current_refresh]];

    //iterate through the control groups dictionary, setting control values to what's in the monitor
    while ((key = [enumerator nextObject])) {
        tmp = [controlGroups objectForKey:key];
        for(i=0; i<[tmp count]; i++) {
            [[tmp objectAtIndex:i] setFloatValue:[myMonitor getHIDElementValue:[key intValue]]];
        }
    }

    //fix the geometry controls separately since the above loop will set them to the last geometry control
    //available instead of the one selected
    [geometrySlider setTag:[[geometrySelector selectedCell] tag]];
    [geometryStepper setTag:[[geometrySelector selectedCell] tag]];
    [geometryValue setTag:[[geometrySelector selectedCell] tag]];    
    [geometrySlider setFloatValue:[myMonitor getHIDElementValue:[[geometrySelector selectedCell] tag]]];
    [geometryStepper setFloatValue:[myMonitor getHIDElementValue:[[geometrySelector selectedCell] tag]]];
    [geometryValue setFloatValue:[myMonitor getHIDElementValue:[[geometrySelector selectedCell] tag]]];
}

//set up the tags associated with the control elements so we know what value they manipulate within the monitor.
//see initControlGroups below for more info: it sets up the controlGroups dictionary that this reads for its info
- (void)initControlTags
{
    int i;
    id key;
    NSArray *tmp;
    NSEnumerator *enumerator = [controlGroups keyEnumerator];

    while ((key = [enumerator nextObject])) {
        tmp = [controlGroups objectForKey:key];
        for(i=0; i<[tmp count]; i++) {
            [[tmp objectAtIndex:i] setTag:[key intValue]];
        }
    }
    
    //geometry tab
    for(i=0; i<[geometrySelector numberOfRows]; i++) {
        [[geometrySelector cellAtRow:i column:0] setTag:geometryElementTags[i]];
    }
}

//set up the controlGroups dictionary, which holds associations between groups of controls and the monitor control they manipulate
//basically, each monitor control has three UI controls which effect it: a slider, a stepper, and a value text field
//this dictionary effectively groups them, keyed by the HID key that identifies that control in the monitor, so we can have
// generic methods for manipulating the controls instead of having a special selector for each one and a *lot* of duplicate code
- (void) initControlGroups
{
    int i;
    controlGroups = [[NSMutableDictionary dictionaryWithCapacity:14] retain]; //value is arbitrary here (just a sizing hint)...
    [controlGroups setObject:[NSArray arrayWithObjects: brightnessSlider,brightnessStepper,brightnessValue,nil] forKey:[NSNumber numberWithInt:kHIDVESAUsagePage_Brightness]];
    [controlGroups setObject:[NSArray arrayWithObjects: contrastSlider,contrastStepper,contrastValue,nil] forKey:[NSNumber numberWithInt:kHIDVESAUsagePage_Contrast]];
    [controlGroups setObject:[NSArray arrayWithObjects: horizontalSizeSlider,horizontalSizeStepper,horizontalSizeValue,nil]
                        forKey:[NSNumber numberWithInt:kHIDVESAUsagePage_HorizSize]];
    [controlGroups setObject:[NSArray arrayWithObjects: horizontalPositionSlider,horizontalPositionStepper,horizontalPositionValue,nil]
                        forKey:[NSNumber numberWithInt:kHIDVESAUsagePage_HorizPosition]];
    [controlGroups setObject:[NSArray arrayWithObjects: verticalSizeSlider,verticalSizeStepper,verticalSizeValue,nil]
                        forKey:[NSNumber numberWithInt:kHIDVESAUsagePage_VertSize]];
    [controlGroups setObject:[NSArray arrayWithObjects: verticalPositionSlider,verticalPositionStepper,verticalPositionValue,nil]
                        forKey:[NSNumber numberWithInt:kHIDVESAUsagePage_VertPosition]];
    [controlGroups setObject:[NSArray arrayWithObjects: redGainSlider,redGainStepper,redGainValue,nil]
                        forKey:[NSNumber numberWithInt:kHIDVESAUsagePage_RedGain]];
    [controlGroups setObject:[NSArray arrayWithObjects: greenGainSlider,greenGainStepper,greenGainValue,nil]
                        forKey:[NSNumber numberWithInt:kHIDVESAUsagePage_GreenGain]];
    [controlGroups setObject:[NSArray arrayWithObjects: blueGainSlider,blueGainStepper,blueGainValue,nil]
                        forKey:[NSNumber numberWithInt:kHIDVESAUsagePage_BlueGain]];
    [controlGroups setObject:[NSArray arrayWithObjects: redLuminanceSlider,redLuminanceStepper,redLuminanceValue,nil]
                        forKey:[NSNumber numberWithInt:kHIDVESAUsagePage_RedLuminance]];
    [controlGroups setObject:[NSArray arrayWithObjects: greenLuminanceSlider,greenLuminanceStepper,greenLuminanceValue,nil]
                        forKey:[NSNumber numberWithInt:kHIDVESAUsagePage_GreenLuminance]];
    [controlGroups setObject:[NSArray arrayWithObjects: blueLuminanceSlider,blueLuminanceStepper,blueLuminanceValue,nil]
                        forKey:[NSNumber numberWithInt:kHIDVESAUsagePage_BlueLuminance]];
    [controlGroups setObject:[NSArray arrayWithObjects: horizontalMisconvergenceSlider,horizontalMisconvergenceStepper,horizontalMisconvergenceValue,nil]
                        forKey:[NSNumber numberWithInt:kHIDVESAUsagePage_HorizMisConv]];
    [controlGroups setObject:[NSArray arrayWithObjects: verticalMisconvergenceSlider,verticalMisconvergenceStepper,verticalMisconvergenceValue,nil]
                        forKey:[NSNumber numberWithInt:kHIDVESAUsagePage_VertMisConv]];
    for(i=0; i<[geometrySelector numberOfRows]; i++) {
        [controlGroups setObject:[NSArray arrayWithObjects: geometrySlider,geometryStepper,geometryValue,nil]
                          forKey:[NSNumber numberWithInt:geometryElementTags[i]]];
    }
}

//set UI elements that are associated with controls that aren't implemented in the monitor or made available by the USB interface
// to disabled to reflect this fact
- (void) disableUnimplementedControls
{
    int i;
    id key;
    NSArray *tmp;
    NSEnumerator *enumerator = [controlGroups keyEnumerator];
    BOOL someEnabled = FALSE;
    int firstEnabledIndex = -1;

    //handle most sliders and associated controls
    //iterate through the control groups setting things based on whether their associated control is available in the monitor
    while ((key = [enumerator nextObject])) {
        if(![myMonitor hasHIDElementWithKey:[key intValue]])
        {
            tmp = [controlGroups objectForKey:key];
            for(i=0; i<[tmp count]; i++) {
                [[tmp objectAtIndex:i] setEnabled:FALSE];
            }            
        }
    }

    //handle disabling degauss button
    if(![myMonitor hasDegauss]) {
        [degaussButton setEnabled:FALSE];
    }
    
    //handle geometry radio controls
    for(i=0; i<[geometrySelector numberOfRows]; i++) {
        if([myMonitor hasHIDElementWithKey:geometryElementTags[i]]) {
            [[geometrySelector cellAtRow:i column:0] setEnabled:TRUE];
            someEnabled=TRUE;
            if(firstEnabledIndex == -1) firstEnabledIndex = i;
        } else {
            [[geometrySelector cellAtRow:i column:0] setEnabled:FALSE];
        }
    }

    //handle the common geometry controls to enable/disable them based on whether any of the radio buttons are active
    if(someEnabled) {
        [geometrySlider setEnabled:TRUE];
        [geometryStepper setEnabled:TRUE];
        [geometryValue setEnabled:TRUE];
        [geometrySelector selectCellAtRow:firstEnabledIndex column:0]; //make the first one selected
    } else {
        [geometrySlider setEnabled:FALSE];
        [geometryStepper setEnabled:FALSE];
        [geometryValue setEnabled:FALSE];
        [geometrySelector selectCellAtRow:0 column:0]; //make the first one selected, even though it's disabled, just to look 'good'
    }
}

#pragma mark -------NSApplication Delegate methods------------
//NSApplication delegate method activated when display params are changed
- (void)applicationDidChangeScreenParameters:(NSNotification *)aNotification
{
    long width, height, refresh;
    CFDictionaryRef newMode = NULL;
    CFNumberRef number = NULL;

    //get the height,width, and refresh for the new mode (maybe I should get this using the NSScreen stuff in Cocoa?)
    if( (newMode = CGDisplayCurrentMode(kCGDirectMainDisplay)) != NULL) {
        number = CFDictionaryGetValue(newMode, kCGDisplayWidth);
        CFNumberGetValue( number, kCFNumberLongType, &width ) ;

        number =  CFDictionaryGetValue(newMode, kCGDisplayHeight);
        CFNumberGetValue( number, kCFNumberLongType, &height ) ;

        number = CFDictionaryGetValue(newMode, kCGDisplayRefreshRate);
        CFNumberGetValue( number, kCFNumberLongType, &refresh ) ;
    } else {
        return;
    }

    //if width, height, or refresh rate changed, get the new settings to revert to and update GUI
    if( (width!=current_width) || (height!=current_height) || (refresh!=current_refresh) ) {
        sleep(1); //wait for monitor settings to stabilize before re-reading them

        //free the old settings before replacing them
        free((void*)originalSettings);

        //grab the settings for this mode so we can revert to them
        originalSettings = [myMonitor settingsArray];

        //update current parameters
        current_width=width;
        current_height=height;
        current_refresh=refresh;
        
        //finally, update the GUI with the current settings
        [self updateControlValues];
    }
}

//NSApplication delegate method to set up our state on launch
- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
    NSMenu* subMenu;
    NSMenuItem* item;
    int numMonitors,i;
    CFDictionaryRef newMode = NULL;
    CFNumberRef number = NULL;

    myMonitor = [[USBMonitor alloc] init];
    if(myMonitor == nil) //suitable device not found
    {
        //window must be in the front to attach a sheet to it for some stupid reason (document THAT, Apple)
        [mainWindow makeKeyAndOrderFront:self]; //TESTME: I wonder if I don't have to do this now?
        NSBeginCriticalAlertSheet(nil, @"Quit", nil, nil, mainWindow, self, @selector(noMonitorSheetDidEnd:returnCode:contextInfo:), NULL, nil, @"No suitable monitor was found.");
    }

    //initialize current display settings
    if( (newMode = CGDisplayCurrentMode(kCGDirectMainDisplay)) != NULL) {
        number = CFDictionaryGetValue(newMode, kCGDisplayWidth);
        CFNumberGetValue( number, kCFNumberLongType, &current_width ) ;

        number =  CFDictionaryGetValue(newMode, kCGDisplayHeight);
        CFNumberGetValue( number, kCFNumberLongType, &current_height ) ;

        number = CFDictionaryGetValue(newMode, kCGDisplayRefreshRate);
        CFNumberGetValue( number, kCFNumberLongType, &current_refresh ) ;
    }    
    
    //set up monitor selector submenu
    subMenu = [[NSMenu alloc] initWithTitle:@"Select"];
    [subMenu setAutoenablesItems:NO]; //we want all items here enabled all the time
    numMonitors = [myMonitor getNumberOfMonitors];
    for(i=0; i<numMonitors; i++)
    {
        item = [subMenu addItemWithTitle:[NSString stringWithFormat:@"%d",(i+1)] action:(SEL)@selector(changeTargetMonitor:) keyEquivalent:@""];
        [item setTag:i];
        if(i==0) //first item is by default the active one
        {
            [item setState:NSOnState];  
        }
    }
    [targetMonitorSelectMenu setSubmenu:subMenu]; //add the submenu

    //get settings as they are before user mucks with them so they can revert if desired
    originalSettings = [myMonitor settingsArray];    

    //set up the array containing associated UI elements
    [self initControlGroups];

    //set up the tag associated with each control so we know which USB element it manipulates
    [self initControlTags];

    //disable controls that this monitor does not implement
    [self disableUnimplementedControls];

    //set up sliders with current values from monitor
    [self updateControlValues];
}

//dispose of things. Done here since the typical -(void)dealloc method doesn't seem to ever get called
- (void)applicationWillTerminate:(NSNotification *)aNotification
{
    free((void*)originalSettings);
    [myMonitor dealloc];
    [super dealloc];
}


#pragma mark -------NSWindow Delegate methods------------
- (BOOL)windowShouldClose:(id)sender
{
    [hideShowControlsWindowMenuItem setTitle:@"Show Controls Window"];
    [self hideTestPatterns:self]; //make sure the test patterns go away when the window closes (don't want the user to get stuck!)
    return YES;
}

@end
